/* Arduino Circuit breaker monitoring system
 * Version 0.5a/Dec 2015 
 */

// Ethernet 
#include <SPI.h>
#include <Ethernet.h>
#include <EthernetUdp.h> // support of UDP packets for NTP time sync
// SD Card
//!!!
// SdFatConfig.h configuration changed to #define SD_SPI_CONFIGURATION 1 
//!!!
#include <SdFat.h>
#include <SdFatUtil.h>
// Time subsystem + Due RTC
#include <Time.h> // Global time object
#include <rtc_clock.h> // DUE on-board RTC
//Experimental
#include <DueTimer.h>
//#include <DS1307RTC.h>
//1-Wire, Power monitoring with INA219 shield
#include <Wire.h>
#include <Adafruit_INA219.h>

// Ini Files 
#include <IniFile.h>


/*-----------------------------------------------------------------
 * Global variables 
  -----------------------------------------------------------------*/
// LED indicators

#define LED_GREEN 48
#define LED_RED 49
#define LED_BLUE 50
#define LED_WHITE 51

// Buttons

#define BUTTON_1 30
#define BUTTON_2 31
#define BUTTON_3 32
#define BUTTON_4 33

// Relais

#define RELAY_1 22
#define RELAY_1 24

// Trigger

#define TRIGGER_IN 40
#define TRIGGER_OUT 41
volatile byte triggerevent = 0;
volatile byte daqdone = 0;

// Watchdog
#define WATCHDOG_OUT 44
#define WATCHDOG_ENABLE 45
volatile byte wdtenable = 0;


// Time functionality

RTC_clock rtc_clock(XTAL);// Due RTC
// Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
const unsigned long seventyYears = 2208988800UL;
unsigned long actual_unixtime;

// NTP Time sync
unsigned int ntp_localPort = 8888;       // local port to listen for UDP packets
char ntp_timeServer[] = "time.nist.gov"; // time.nist.gov NTP server
const int NTP_PACKET_SIZE = 48; // NTP time stamp is in the first 48 bytes of the message
byte ntp_packetBuffer[ NTP_PACKET_SIZE]; //buffer to hold incoming and outgoing packets
EthernetUDP Udp; // A UDP instance to let us send and receive packets over UDP


// Hardware
Adafruit_INA219 ina219; //1-Wire, Power monitoring device INA219 shield

// Ethernet 
#define ETHERNET_BUFFER_SIZE 128
byte mac[] = {0x00, 0xAA, 0xBB, 0xCC, 0xDE, 0x02};
IPAddress ip(192, 168, 2, 101);
byte useStaticIP = 1;
char rootFileName[] = "index.htm";
EthernetServer server(80);

// AJAX
int ajaxrequest = 0;
byte ajaxupdaterequired = 0; //send update required state command for new recorded event after triggering

// SD Card
#define SD_BUFFER_SIZE 2048
byte buff[SD_BUFFER_SIZE];
SdFat sd;
SdFile file;
SdFile root;
SdFile logfile;
SdFile eventlistfile;
SdFile comtradefile;
#define COMTRADE_FILE_BASE_NAME "event"

// Utilites
long begintime;
long endtime;
long Tasktimer = 0;

// Ini Files 
const size_t inibufferLen = 80;
char inibuffer[inibufferLen];
const char *inifilename = "/net.ini";

// Serial streams
ArduinoOutStream cout(Serial);

// System metrics
float SMshuntvoltage = 0;
float SMbusvoltage = 0;
float SMcurrent_mA = 0;
float SMloadvoltage = 0;

 // ADC 
#define ADC_BUFFER_SIZE  4096
#define ADC_SAMPLES_PER_CHANNEL  ADC_BUFFER_SIZE/8
#define ADC_SAMPLERATE  52600 / 8
#define ADC_SAMPLETIME  152  // sample time in us
uint16_t adcbuffer[ADC_BUFFER_SIZE];
long adcstartmicros;
long adcendmicros;
time_t adctime;
long eventsrecordedsincereboot = 0;

// Digital signal processing variables and constants

#define DSP_DIGITAL_TRESHOLD  3500

static int analogscale [4] = {2,2,2,2}; // dividers for analog channels
static int analogoffset [4] = {-2048,-2048,-2048,-2048};// offset for analog channels

static int calcscale [4] = {64,64,64,1}; // dividers for calculated values
static int calcoffset [4] = {1,1,1,0};// offset for calculated values


#define DSP_FIR_FILTERTAPS  15
#define DSP_FIR_DCGAIN  262144

int analogbuffer_ch0[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled values for analog channels
int analogbuffer_ch1[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled values for analog channels
int analogbuffer_ch2[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled values for analog channels
int analogbuffer_ch3[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled values for analog channels

byte digitalbuffer_ch0[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled values for digital channels
byte digitalbuffer_ch1[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled values for digital channels
byte digitalbuffer_ch2[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled values for digital channels
byte digitalbuffer_ch3[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled values for digital channels

int calculatedbuffer_ch0[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled calculated values
int calculatedbuffer_ch1[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled calculated values
int calculatedbuffer_ch2[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled calculated values
int calculatedbuffer_ch3[ADC_SAMPLES_PER_CHANNEL]; // store filtered and scaled calculated values
/**************************************************************
Filter type: Low Pass
Filter model: Raised Cosine
Roll Off Factor: 0.500000
Sampling Frequency: 6 KHz
Cut Frequency: 0.250000 KHz
Coefficents Quantization: 16-bit
***************************************************************/

int analogfiltercoeff[DSP_FIR_FILTERTAPS] = { 
        11368,
        13693,
        15875,
        17819,
        19436,
        20649,
        21401,
        21656,
        21401,
        20649,
        19436,
        17819,
        15875,
        13693,
        11368        
    };

 
/*-----------------------------------------------------------------
 * Setup for all hardware devices and software modules 
  -----------------------------------------------------------------*/

void setup() {

  Serial.begin(9600); 

  cout << F("---------------------------------------------") << endl;
  cout << F("-           CB MONITORING SYSTEM            -") << endl;
  cout << F("-           Version 0.2/23.12.2015          -") << endl;
  cout << F("---------------------------------------------") << endl;
  cout << endl << endl;

  PgmPrint("Free RAM: ");
  Serial.println(FreeRam());
  
  cout << F("-------------------I2C-----------------------") << endl;
  InitI2C(); // Enumerate i2c devices
  cout << F("-------------------RTC-----------------------") << endl;

  rtc_clock.init(); // Init DUE RTC 
 

  ina219.begin(); // Init i2c communication with ina219 shield

// Init digital inputs and outputs
 InitIO();


  // turn blue led on for during system start
  digitalWrite(LED_BLUE, HIGH);  
  InitSDcard();// Init SDcard

  cout << F("-------------------Read settings-------------------") << endl;

  ReadIniFile(); //Read settings from ini file

  InitADCSystem(); // Init ADC

  cout << F("--------------Init ethernet communication----------") << endl;

  InitEthernetSystem();
  
 
  GetNTPTime();  // NTP Request
  rtc_clock.set_clock(actual_unixtime, Germany); // Set system time
  setTime(actual_unixtime);

   setSyncProvider(getArduinoDueTime);
  setSyncInterval(5);
  if(timeStatus()!= timeSet) 
     Serial.println("Unable to sync with the RTC");
  else
     Serial.println("RTC has set the system time");      

  
  time_t ttmp = now();
  
  Serial.println(); 
  Serial.print("System time is: "); 
  Serial.print(hour(ttmp)); 
  Serial.print(":"); 
  Serial.print(minute(ttmp)); 
  Serial.print(":"); 
  Serial.print(second(ttmp)); 
  Serial.println(); 

  
  // start task timer
  

  // selftest daq

  Serial.println("Selftest: ADC conversion in progress....");
  RunADC(); // start DAQ with DMA to adcbuffer  
  ProcessADCBuffer(); // apply digital signal filtering
  Serial.println("Selftest: ADC conversion done!");
  Serial.println("Selftest: Writing comtrade file....");
  WriteComtradeFile();
  Serial.println("Selftest: done!");
  Serial.println();

  Tasktimer = millis();
  cout << F("--------------Main loop------------") << endl;
 // turn blue led off
 digitalWrite(LED_BLUE, LOW);

 //-----------------------Wathdog-----------------------
wdtenable = 1;
digitalWrite(WATCHDOG_ENABLE, HIGH); // enable external watchdog device

}

/*-----------------------------------------------------------------
 * Main loop of the programm 
  -----------------------------------------------------------------*/

void loop() {

  
   
  char clientline[ETHERNET_BUFFER_SIZE];
  char *filename;
  int image = 0;
  int index = 0;


 EthernetClient client = server.available();
  if (client) {
    // an http request ends with a blank line
    Serial.println("new client");
    boolean current_line_is_blank = true;
    index = 0;
//---
// clear client line

for (int i = 0; i < ETHERNET_BUFFER_SIZE; i++) {
        clientline[i] = 0;
    }

//----

        
   while (client.connected()) {
      if (client.available()) {
        char c = client.read();

        if (c != '\n' && c != '\r') {
          clientline[index] = c;
          index++;
          if (index >= ETHERNET_BUFFER_SIZE) 
            index = ETHERNET_BUFFER_SIZE -1;

          continue;
        }
        ajaxrequest = 0;
        clientline[index] = 0;
        filename = 0;
        Serial.println("CLIENT LINE:");
        Serial.println(clientline);
        
        digitalWrite(LED_BLUE, HIGH); //start communication indicator

        //------- parse AJAX request--------------

        if (strstr(clientline, "ajax") != 0) { // client line is an  AJAX request
          ajaxrequest = 1;
          
          if (strstr(clientline, "off") != 0) { // software reset for "off" AJAX command
            digitalWrite(LED_BLUE, HIGH);
            digitalWrite(LED_WHITE, HIGH);
            digitalWrite(LED_RED, HIGH);
            digitalWrite(LED_GREEN, HIGH);
            Serial.println("Resetting.....");
            delay(5000);
            Software_Reset();
          } 

          if (strstr(clientline, "daq") != 0) { // start daq by "daq" AJAX command
            digitalWrite(LED_GREEN, HIGH);
            Serial.println("ADC conversion in progress....");
            RunADC(); // start DAQ with DMA to adcbuffer  
            ProcessADCBuffer(); // apply digital signal filtering
            Serial.println("ADC conversion done!");
            digitalWrite(LED_GREEN, LOW);
            digitalWrite(LED_RED, HIGH);
            Serial.println("Writing comtrade file....");
            WriteComtradeFile();
            Serial.println("Writing comtrade file one!");
            Serial.println();
            digitalWrite(LED_RED, LOW);
            
          } 

          if (strstr(clientline, "removeeventlist") != 0) { // remove eventlist by "removeeventlist" AJAX command
            Serial.println("Deleting event list.....");
            RemoveEventList();
            Serial.print("Done!");
            Serial.println();
            
          } 
          
          Serial.println("AJAX REQUEST FOUND!");
          
          client.println("HTTP/1.1 200 OK");
          client.println("Content-Type: text/xml");
          client.println("Connection: keep-alive");
          client.println();
          SendAJAXRespone(client);

          // clear client line
          for (int i = 0; i < ETHERNET_BUFFER_SIZE; i++) {
            clientline[i] = 0;
          }

        } 
        else // client line is not an AJAX request
       {

        if (strstr(clientline, "GET / ") != 0) {
          filename = rootFileName;
        } 
        
        if (strstr(clientline, "GET /") != 0) {
          if (!filename) filename = clientline + 5; 

          (strstr(clientline, " HTTP"))[0] = 0;

          Serial.println(filename);

          if (! file.open(filename, O_READ)) {
            Serial.println("404 1:");
            client.println("HTTP/1.1 404 Not Found");
            client.println("Content-Type: text/html");
            client.println();
            client.println("<h2>Error 404</h2>");
            client.println("<s2>The file does not exist.<s2>");
            client.println("");
            break;
          }

          Serial.println("Opened!");
          //File types
          client.println("HTTP/1.1 200 OK");
          if (strstr(filename, ".htm") != 0)
            client.println("Content-Type: text/html");
          else if (strstr(filename, ".css") != 0)
            client.println("Content-Type: text/css");
          else if (strstr(filename, ".png") != 0)
            client.println("Content-Type: image/png");
          else if (strstr(filename, ".jpg") != 0)
            client.println("Content-Type: image/jpeg");
          else if (strstr(filename, ".gif") != 0)
            client.println("Content-Type: image/gif");
          else if (strstr(filename, ".3gp") != 0)
            client.println("Content-Type: video/mpeg");
          else if (strstr(filename, ".pdf") != 0)
            client.println("Content-Type: application/pdf");
          else if (strstr(filename, ".js") != 0)
            client.println("Content-Type: application/x-javascript");
          else if (strstr(filename, ".xml") != 0)
            client.println("Content-Type: application/xml");
          else
            client.println("Content-Type: text");
          client.println();

          int16_t c;
          Serial.println ("START...");
          begintime=micros();  
              
//----------------send requested file to client--------------------------
          unsigned long filesize = 0;
          while ((filesize = file.read(buff, sizeof(buff))) > 0) {   
          client.write((uint8_t*)buff, filesize);
         }
//-----------------------------------------------------------------------

        endtime=micros();
        Serial.print("STOP, done in ");
        Serial.print(endtime-begintime);
        Serial.print("micros\n\n");  
        
        file.close();
        } 
        else {
          if (ajaxrequest == 1)
          {
          Serial.println("404 2:");
          client.println("HTTP/1.1 404 Not Found");
          client.println("Content-Type: text/html");
          client.println();
          client.println("<h2>Error 404</h2>");
          client.println("");
          }
        }
        break;
      }
      }
    }
    
    delay(1);
    client.stop();
    Serial.println("client disconnected");
    digitalWrite(LED_BLUE, LOW); // end of communication
  }

// -------------------- primitive task scheduler -------------------------

  long currenttime;
  currenttime = millis();

//-----------------first task, each 3 sec-----------------------
  if ((currenttime - Tasktimer)>3000) // do each 3 sec
  {
    digitalWrite(LED_WHITE, HIGH); // start of task indicator
    
    GetSystemMetrics(); // update state of system variables: voltage, current etc.
    
   // WriteSystemLog(); // write system variables to log file

 //   RunADC(); // start DAQ with DMA to adcbuffer  

 //   WriteComtradeFile();

    cout << "timestatus: " << timeStatus() << endl;

    Tasktimer = millis(); // update task timer
    digitalWrite(LED_WHITE, LOW); // end of task indicator
  } 
//-----------------------end of first task-----------------------
// ... more tasks


//-----------------Triggering-----------------------
  triggerevent = digitalRead(TRIGGER_IN);
  digitalWrite(TRIGGER_OUT, triggerevent);
  
  if ( (triggerevent > 0) && (daqdone == 0) ){

            digitalWrite(LED_GREEN, HIGH);
            Serial.println("ADC conversion in progress....");
            RunADC(); // start DAQ with DMA to adcbuffer  
            ProcessADCBuffer(); // apply digital signal filtering
            Serial.println("ADC conversion done!");
            digitalWrite(LED_GREEN, LOW);
            digitalWrite(LED_RED, HIGH);
            Serial.println("Writing comtrade file....");
            WriteComtradeFile();
            Serial.println("Writing comtrade file one!");
            Serial.println();
            digitalWrite(LED_RED, LOW);
            daqdone = 1;
            ajaxupdaterequired = 1;
    
  }

   if ( (triggerevent == 0) && (daqdone > 0) ){
    daqdone = 0;
   }

  
  
//-----------------------end triggering-----------------------

//-----------------------wathdog-----------------------

digitalWrite(WATCHDOG_OUT, !digitalRead(WATCHDOG_OUT)); // send pulse to external watchdog device

}


/*-----------------------------------------------------------------
 * Support functions
  -----------------------------------------------------------------*/

void GetSystemMetrics()
{
// ina
  SMshuntvoltage = ina219.getShuntVoltage_mV();
  SMbusvoltage = ina219.getBusVoltage_V();
  SMcurrent_mA = ina219.getCurrent_mA();
  SMloadvoltage = SMbusvoltage + (SMshuntvoltage / 1000);  
}


void SendAJAXRespone(EthernetClient cl)
{    
    cl.print("<?xml version = \"1.0\" ?>");
    cl.print("<states>");
    cl.print("<state1>");
    cl.print(SMbusvoltage);   
    cl.print("</state1>");
    cl.print("<state2>");   
    cl.print(millis());    
    cl.print("</state2>");   
    cl.print("<state3>");
    cl.print(FreeRam());
    cl.print("</state3>");
    cl.print("<state4>");
    cl.print(hour());
    cl.print(":");
    cl.print(minute());
    cl.print(":");
    cl.print(second());
    cl.print("</state4>");

    cl.print("<state5>");
      for (byte thisByte = 0; thisByte < 4; thisByte++) {
    // print the value of each byte of the IP address:
    cl.print(Ethernet.localIP()[thisByte], DEC);
    cl.print(".");
  }   
    cl.print("</state5>");

    cl.print("<state6>");
    cl.print(ajaxupdaterequired);  
    if (ajaxupdaterequired > 0){      
      ajaxupdaterequired = 0; 
    }
    cl.print("</state6>");

    cl.print("<state7>");
    cl.print(eventsrecordedsincereboot);  
    cl.print("</state7>");

    
    
    cl.print("</states>");

    
}


time_t getArduinoDueTime()
{
  return rtc_clock.unixtime();
}


//------------------------------------------------------------------------------
//
void fatalBlink() {
  while (true) {   
      digitalWrite(LED_WHITE, HIGH);
      digitalWrite(LED_BLUE, HIGH);
      digitalWrite(LED_RED, HIGH);
      digitalWrite(LED_GREEN, HIGH);
      delay(200);
      digitalWrite(LED_WHITE, LOW);
      digitalWrite(LED_BLUE, LOW);
      digitalWrite(LED_RED, LOW);
      digitalWrite(LED_GREEN, LOW);
      delay(200);
  }
}


// String utility
String getStringPartByNr(String data, char separator, int index)
{
    // spliting a string and return the part nr index
    // split by separator
    
    int stringData = 0;        //variable to count data part nr 
    String dataPart = "";      //variable to hole the return text
    
    for(int i = 0; i<=data.length()-1; i++) {    //Walk through the text one letter at a time
      
      if(data[i]==separator) {
        //Count the number of times separator character appears in the text
        stringData++;
        
      }else if(stringData==index) {
        //get the text when separator is the rignt one
        dataPart.concat(data[i]);
        
      }else if(stringData>index) {
        //return text and stop if the next separator appears - to save CPU-time
        return dataPart;
        break;
        
      }

    }
    //return text if this is the last part
    return dataPart;
}



void InitI2C() {

  byte iic_error, iic_address;
  int iic_nDevices;

  Wire.begin();
  cout << F("\nScanning for iic devices...\n");

  iic_nDevices = 0;
  for(iic_address = 1; iic_address < 127; iic_address++ ) 
  {
    // The i2c_scanner uses the return value of
    // the Write.endTransmisstion to see if
    // a device did acknowledge to the address.
    Wire.beginTransmission(iic_address);
    iic_error = Wire.endTransmission();

    if (iic_error == 0)
    {
      cout << F("\nI2C device found at address 0x\n");
      if (iic_address<16) 
        Serial.print("0");
      Serial.print(iic_address,HEX);
      Serial.println("  !");

      iic_nDevices++;
    }
    else if (iic_error==4) 
    {
      Serial.print("Unknow error at address 0x");
      if (iic_address<16) 
        Serial.print("0");
      Serial.println(iic_address,HEX);
    }    
  }
  if (iic_nDevices == 0)
    Serial.println("No I2C devices found\n");
  else
    Serial.println("done\n");  
}

/*-----------------------------------------------------------------
 * Software reset 
  -----------------------------------------------------------------*/

void Software_Reset() {
  const int RSTC_KEY = 0xA5;
  RSTC->RSTC_CR = RSTC_CR_KEY(RSTC_KEY) | RSTC_CR_PROCRST | RSTC_CR_PERRST;
  while (true);
}

/*-----------------------------------------------------------------
 * Init input / output pins 
  -----------------------------------------------------------------*/

void InitIO() {
 // Init digital outputs for LED indicators
  pinMode(LED_BLUE, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_WHITE, OUTPUT);
  pinMode(LED_RED, OUTPUT);
  
  digitalWrite(LED_BLUE, LOW);    
  digitalWrite(LED_GREEN, LOW);   
  digitalWrite(LED_WHITE, LOW);   
  digitalWrite(LED_RED, LOW);   

  pinMode(TRIGGER_IN, INPUT);

  pinMode(TRIGGER_OUT, OUTPUT);
  digitalWrite(TRIGGER_OUT, LOW);

  pinMode(BUTTON_1, INPUT);
  pinMode(BUTTON_2, INPUT);
  pinMode(BUTTON_3, INPUT);
  pinMode(BUTTON_4, INPUT);

  digitalWrite(BUTTON_1, HIGH);  
  digitalWrite(BUTTON_2, HIGH); 
  digitalWrite(BUTTON_3, HIGH); 
  digitalWrite(BUTTON_4, HIGH);  

  pinMode(WATCHDOG_OUT, OUTPUT);
  digitalWrite(WATCHDOG_OUT, LOW);

   pinMode(WATCHDOG_ENABLE, OUTPUT);
  digitalWrite(WATCHDOG_ENABLE, LOW);
}

/*-----------------------------------------------------------------
 * NTP Time sync functions 
  -----------------------------------------------------------------*/


void GetNTPTime() {


  Udp.begin(ntp_localPort);
  sendNTPpacket(ntp_timeServer); // send an NTP packet to a time server
  cout << F("\nsend an NTP packet to a time server.\n");

  // wait to see if a reply is available
  delay(1000);
  if (Udp.parsePacket()) {
    cout << F("\nWe've received a packet.\n");
    // We've received a packet, read the data from it
    Udp.read(ntp_packetBuffer, NTP_PACKET_SIZE); // read the packet into the buffer

    //the timestamp starts at byte 40 of the received packet and is four bytes,
    // or two words, long. First, esxtract the two words:

    unsigned long highWord = word(ntp_packetBuffer[40], ntp_packetBuffer[41]);
    unsigned long lowWord = word(ntp_packetBuffer[42], ntp_packetBuffer[43]);
    // combine the four bytes (two words) into a long integer
    // this is NTP time (seconds since Jan 1 1900):
    unsigned long secsSince1900 = highWord << 16 | lowWord;
    actual_unixtime = secsSince1900 - seventyYears;    
    Serial.print("Seconds since Jan 1 1900 = ");
    Serial.println(secsSince1900);
/*
    // now convert NTP time into everyday time:
    Serial.print("Unix time = ");
    // Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
    const unsigned long seventyYears = 2208988800UL;
    // subtract seventy years:
    unsigned long epoch = secsSince1900 - seventyYears;
    // print Unix time:
    Serial.println(epoch);


    // print the hour, minute and second:
    Serial.print("The UTC time is ");       // UTC is the time at Greenwich Meridian (GMT)
    Serial.print((epoch  % 86400L) / 3600); // print the hour (86400 equals secs per day)
    Serial.print(':');
    if (((epoch % 3600) / 60) < 10) {
      // In the first 10 minutes of each hour, we'll want a leading '0'
      Serial.print('0');
    }
    Serial.print((epoch  % 3600) / 60); // print the minute (3600 equals secs per minute)
    Serial.print(':');
    if ((epoch % 60) < 10) {
      // In the first 10 seconds of each minute, we'll want a leading '0'
      Serial.print('0');
    }
    Serial.println(epoch % 60); // print the second
    */
  }
  // wait ten seconds before asking for the time again
 // delay(10000);
 // Ethernet.maintain();

  
}

// send an NTP request to the time server at the given address
unsigned long sendNTPpacket(char* address) {
  // set all bytes in the buffer to 0
  memset(ntp_packetBuffer, 0, NTP_PACKET_SIZE);
  // Initialize values needed to form NTP request
  // (see URL above for details on the packets)
  ntp_packetBuffer[0] = 0b11100011;   // LI, Version, Mode
  ntp_packetBuffer[1] = 0;     // Stratum, or type of clock
  ntp_packetBuffer[2] = 6;     // Polling Interval
  ntp_packetBuffer[3] = 0xEC;  // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  ntp_packetBuffer[12]  = 49;
  ntp_packetBuffer[13]  = 0x4E;
  ntp_packetBuffer[14]  = 49;
  ntp_packetBuffer[15]  = 52;

  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp:
  Udp.beginPacket(address, 123); //NTP requests are to port 123
  Udp.write(ntp_packetBuffer, NTP_PACKET_SIZE);
  Udp.endPacket();
}

/*-----------------------------------------------------------------
 * SD card functions 
  -----------------------------------------------------------------*/

void InitSDcard() {
  pinMode(10, OUTPUT);                       // set the SS pin as an output (necessary!)
  digitalWrite(10, HIGH);                    // but turn off the W5100 chip!

if (!sd.begin(4, SPI_FULL_SPEED)) {
    if (sd.card()->errorCode()) {
      cout << F(
             "\nSD initialization failed.\n"
             "Do not reformat the card!\n"
             "Is the card correctly inserted?\n"
             "Is chipSelect set to the correct value?\n"
             "Does another SPI device need to be disabled?\n"
             "Is there a wiring/soldering problem?\n");
      cout << F("\nerrorCode: ") << hex << showbase;
      cout << int(sd.card()->errorCode());
      cout << F(", errorData: ") << int(sd.card()->errorData());
      cout << dec << noshowbase << endl;
      fatalBlink();
      return;
    }
    cout << F("\nCard successfully initialized.\n");
    if (sd.vol()->fatType() == 0) {
      cout << F("Can't find a valid FAT16/FAT32 partition.\n");
      return;
    }
    if (!sd.vwd()->isOpen()) {
      cout << F("Can't open root directory.\n");

      return;
    }
    cout << F("Can't determine error type\n");
    return;
  }
  cout << F("\nCard successfully initialized.\n");
  cout << endl;

 uint32_t size = sd.card()->cardSize();
  if (size == 0) {
    cout << F("Can't determine the card size.\n");
   
    return;
  }
  uint32_t sizeMB = 0.000512 * size + 0.5;
  cout << F("Card size: ") << sizeMB;
  cout << F(" MB (MB = 1,000,000 bytes)\n");
  cout << endl;
  cout << F("Volume is FAT") << int(sd.vol()->fatType());
  cout << F(", Cluster size (bytes): ") << 512L * sd.vol()->blocksPerCluster();
  cout << endl << endl;

  cout << F("Files found (date time size name):\n");
  sd.ls(LS_R | LS_DATE | LS_SIZE);

  if ((sizeMB > 1100 && sd.vol()->blocksPerCluster() < 64)
      || (sizeMB < 2200 && sd.vol()->fatType() == 32)) {
    cout << F("\nThis card should be reformatted for best performance.\n");
    cout << F("Use a cluster size of 32 KB for cards larger than 1 GB.\n");
    cout << F("Only cards larger than 2 GB should be formatted FAT32.\n");
    return;
  }

    
}
/*-----------------------------------------------------------------
 * Ini file functions 
  -----------------------------------------------------------------*/

void ReadIniFile() {

 IniFile ini(inifilename);
  if (!ini.open()) {
    Serial.print("Ini file ");
    Serial.print(inifilename);
    Serial.println(" does not exist");
    // Cannot do anything else
    while (1)
      ;
  }
  Serial.println("Ini file exists");

   if (!ini.validate(inibuffer, inibufferLen)) {
    Serial.print("ini file ");
    Serial.print(ini.getFilename());
    Serial.print(" not valid: ");
    // Cannot do anything else
    while (1)
      ;
  }

   if (ini.getValue("network", "usestaticip", inibuffer, inibufferLen)) {
    Serial.print("section 'network' has an entry 'usestaticip' with value ");
    Serial.println(inibuffer);
    int val = atoi(inibuffer);
    
    if (val == 1) {
      useStaticIP = 1;

      }
      else {
        useStaticIP = 0;

        }
        
  }
  else {
    Serial.print("Could not read 'usestaticip' from section 'network', use default value ");
  }

  if (ini.getValue("network", "ip", inibuffer, inibufferLen)) {
    Serial.print("section 'network' has an entry 'ip' with value ");
    Serial.println(inibuffer);
    String ipAdress = "";
    ipAdress= getStringPartByNr(inibuffer,'.',0);
    ip[0] = ipAdress.toInt();
    ipAdress= getStringPartByNr(inibuffer,'.',1);
    ip[1] = ipAdress.toInt();
    ipAdress= getStringPartByNr(inibuffer,'.',2);
    ip[2] = ipAdress.toInt();
    ipAdress= getStringPartByNr(inibuffer,'.',3);
    ip[3] = ipAdress.toInt();    
    Serial.println(ip);    
  }
  else {
    Serial.print("Could not read 'ip' from section 'network', use default value ");
  }

}

  /*-----------------------------------------------------------------
 * systemlog file functions 
  -----------------------------------------------------------------*/

void WriteSystemLog(){

if (logfile.open("systemlog.txt", O_RDWR | O_CREAT | O_AT_END)) {
    Serial.println("opening systemlog.txt for write");
    logfile.println();
    logfile.print(hour());
    logfile.print(":");
    logfile.print(minute());
    logfile.print(":");
    logfile.print(second()); 
    logfile.print(" uptime:");   
    logfile.print(micros());
    logfile.print("ms, bus voltage:");   
    logfile.print(SMbusvoltage);
    logfile.print(" v, events since reboot:");    
    logfile.print(eventsrecordedsincereboot);
    logfile.close();
    
  }

  
}


 /*-----------------------------------------------------------------
 * Comtrade functions 
  -----------------------------------------------------------------*/


  void RemoveEventList ()

  {

    sd.remove("eventlist.txt");
    if (eventlistfile.open("eventlist.txt", O_RDWR | O_CREAT | O_AT_END)) {
  Serial.println("event list file recreated");
  eventlistfile.close(); 
}
    
  }

  void WriteComtradeFile()
  {

    // 1 more event recorded

    eventsrecordedsincereboot++;
    
    // generate file name
  String cfgfileName = COMTRADE_FILE_BASE_NAME; 
  String datfileName = COMTRADE_FILE_BASE_NAME; 
  String eventDateTime = "";
  
  cfgfileName = cfgfileName + "_" + year(adctime) + "."+ month(adctime) +"."+ day (adctime)+ "_"+ hour(adctime) + "." + minute (adctime) + "." + second (adctime) +"_" + adcstartmicros+".cfg";
  datfileName = datfileName + "_" + year(adctime) + "."+ month(adctime) +"."+ day (adctime)+ "_"+ hour(adctime) + "." + minute (adctime) + "." + second (adctime) +"_" + adcstartmicros+".dat";
  Serial.println(cfgfileName);

  // write event to eventlist

       // convert string to char array
  int str_len = datfileName.length() + 1; 
  char charfileName[str_len];
  datfileName.toCharArray(charfileName, str_len);

if (eventlistfile.open("eventlist.txt", O_RDWR | O_CREAT | O_AT_END)) {
  Serial.println("opening event list file for write");
  eventlistfile.print(eventsrecordedsincereboot);
  eventlistfile.print(",");
  eventlistfile.print(charfileName);
  eventlistfile.println();
  eventlistfile.close(); 
}
  

    //write CFG file

     // convert string to char array
   str_len = cfgfileName.length() + 1; 
   charfileName[str_len];
  cfgfileName.toCharArray(charfileName, str_len);
    
if (comtradefile.open(charfileName, O_RDWR | O_CREAT)) {
    Serial.println("opening cfg event file for write");

    comtradefile.println("test station, 1");
    comtradefile.println("8,8A,4D");
    comtradefile.println("1,ACH1,PH1,, A,  1.0,0.0,0.0,-1024,1024");
    comtradefile.println("2,ACH2,PH2,, A,  1.0,0.0,0.0,-1024,1024");
    comtradefile.println("3,ACH3,PH3,, A,  1.0,0.0,0.0,-1024,1024");
    comtradefile.println("4,ACH4,S,, mm,  1.0,0.0,0.0,0,1024");
    comtradefile.println("5,CCH1,PH1,, A2s,  1.0,0.0,0.0,-1024,1024");
    comtradefile.println("6,CCH2,PH2,, A2s,  1.0,0.0,0.0,-1024,1024");
    comtradefile.println("7,CCH3,PH3,, A2s,  1.0,0.0,0.0,-1024,1024");
    comtradefile.println("8,CCH4,S,, mm,  1.0,0.0,0.0,-1024,1024");
    comtradefile.println("1,DCH1,0");
    comtradefile.println("2,DCH2,0");
    comtradefile.println("3,DCH3,0");
    comtradefile.println("4,DCH4,0");
    comtradefile.println("50");
    comtradefile.println("1");
    comtradefile.println("52600.00,512");
    
    eventDateTime = eventDateTime + day (adctime) + "/" + month(adctime)+ "/" + year(adctime)+ ", " + hour(adctime) + ":" + minute (adctime) + ":" + second (adctime)+ ". 0";
    
    comtradefile.println(eventDateTime);
    comtradefile.println(eventDateTime);
    comtradefile.println("ASCII");  
     
    comtradefile.close();    
 }

    //write DAT file


     // convert string to char array
  str_len = datfileName.length() + 1; 
  charfileName[str_len];
  datfileName.toCharArray(charfileName, str_len);

  uint16_t tempadcvalue;
    
if (comtradefile.open(charfileName, O_RDWR | O_CREAT)) {
    Serial.println("opening dat event file for write");

    int samplenumber;
    for (int i=0; i<ADC_SAMPLES_PER_CHANNEL; i++)
  {

    comtradefile.print(i+1); // sample number 1,2,3,....
    comtradefile.print(",");
    
    comtradefile.print(i*152); // 152 us for each sample. time
    comtradefile.print(",");

    // analog channels
    comtradefile.print(analogbuffer_ch0[i]);
    comtradefile.print(",");

    comtradefile.print(analogbuffer_ch1[i]);
    comtradefile.print(",");

    comtradefile.print(analogbuffer_ch2[i]);
    comtradefile.print(",");

    comtradefile.print(analogbuffer_ch3[i]);
    comtradefile.print(",");

    // calculated channels
    comtradefile.print(calculatedbuffer_ch0[i]);
    comtradefile.print(",");

    comtradefile.print(calculatedbuffer_ch1[i]);
    comtradefile.print(",");

    comtradefile.print(calculatedbuffer_ch2[i]);
    comtradefile.print(",");

    comtradefile.print(calculatedbuffer_ch3[i]);
    comtradefile.print(",");
    
    // digital channels
    comtradefile.print(digitalbuffer_ch0[i]);
    comtradefile.print(",");

    comtradefile.print(digitalbuffer_ch1[i]);
    comtradefile.print(",");

    comtradefile.print(digitalbuffer_ch2[i]);
    comtradefile.print(",");

    comtradefile.print(digitalbuffer_ch3[i]);
    comtradefile.print(",");  
   
    comtradefile.println();
    
  
 }

   
     
    comtradefile.close();    
 }
    
    
    
    
    
   
    
}

  


 /*-----------------------------------------------------------------
 * ADC functions 
  -----------------------------------------------------------------*/

void InitADCSystem()
{
  // ADC SETUP

//pinMode(ainput,INPUT);

//adc_init(ADC, sysclk_get_cpu_hz(), ADC_FREQ_MIN, ADC_STARTUP_TIME_4);
//adc_configure_timing(ADC, TRACKING_TIME, ADC_SETTLING_TIME_3, TRANSFER_PERIOD);

/*
ADC_FREQ_MIN = 1000000 (1Mhz)
TRACKING_TIME = 1
ADC_SETTLING_TIME_3 = 17
TRANSFER_PERIOD = 1

Sampling rate =  ((TRACKING_TIME * 2 + 3) + (TRANSFER_PERIOD + 1) + ADC_SETTLING_TIME_3) / ADC_FREQ_MIN = (1 + 1 + 17)/1000000 = 52.6Khz


*/

/*
 * temp sensor

 adc_enable_channel(ADC, ADC_TEMPERATURE_SENSOR);
  adc_enable_ts(ADC);
 */

pinMode(A0, OUTPUT);
digitalWrite(A0, LOW);

pinMode(A1, OUTPUT);
digitalWrite(A1, LOW);

pinMode(A2, OUTPUT);
digitalWrite(A2, LOW);

pinMode(A3, OUTPUT);
digitalWrite(A3, LOW);

pinMode(A4, OUTPUT);
digitalWrite(A4, LOW);

pinMode(A5, OUTPUT);
digitalWrite(A5, LOW);

pinMode(A6, OUTPUT);
digitalWrite(A6, LOW);

pinMode(A7, OUTPUT);
digitalWrite(A7, LOW);

  // Setup all registers
  pmc_enable_periph_clk (ID_ADC);
  adc_init (ADC, SystemCoreClock, ADC_FREQ_MIN, ADC_STARTUP_FAST);
  adc_disable_interrupt (ADC, 0xFFFFFFFF);
  adc_set_resolution (ADC, ADC_10_BITS);
  adc_configure_power_save (ADC, ADC_MR_SLEEP_NORMAL, ADC_MR_FWUP_OFF);
  //adc_configure_timing(ADC, TRACKING_TIME, ADC_SETTLING_TIME_3, TRANSFER_PERIOD);
  adc_configure_timing (ADC, 1, ADC_SETTLING_TIME_3, 1);
  adc_set_bias_current (ADC, 1);
  adc_disable_tag (ADC);
  adc_disable_ts (ADC);
  adc_stop_sequencer (ADC);
  adc_disable_channel_differential_input (ADC, ADC_CHANNEL_7);
  adc_disable_channel_differential_input (ADC, ADC_CHANNEL_5);
  adc_disable_channel_differential_input (ADC, ADC_CHANNEL_3);
  adc_disable_channel_differential_input (ADC, ADC_CHANNEL_1);
  adc_disable_all_channel (ADC);
  adc_enable_channel (ADC, ADC_CHANNEL_7); // A0
  adc_enable_channel (ADC, ADC_CHANNEL_6); // A1
  adc_enable_channel (ADC, ADC_CHANNEL_5); // A2
  adc_enable_channel (ADC, ADC_CHANNEL_4); // A3
  adc_enable_channel (ADC, ADC_CHANNEL_3); // A4
  adc_enable_channel (ADC, ADC_CHANNEL_2); // A5
  adc_enable_channel (ADC, ADC_CHANNEL_1); // A6
  adc_enable_channel (ADC, ADC_CHANNEL_0); // A7

  //-------------------------------------------------------
 // adc_enable_channel(ADC, ADC_TEMPERATURE_SENSOR);
 // adc_enable_ts(ADC);
  //-------------------------------------------------------


// adc_channel_num_t channelssequence[] = {ADC_CHANNEL_7, ADC_CHANNEL_5, ADC_CHANNEL_3, ADC_TEMPERATURE_SENSOR};

//  adc_configure_sequence(ADC, channelssequence, 4);
//  adc_start_sequencer (ADC);
// !!!!! conversion results are 16 bit values 4MSB-channel number, 12 LSB- value : ccccrrrrrrrrrrrr
  adc_enable_tag (ADC);
  
  adc_configure_trigger (ADC, ADC_TRIG_SW, 1);
  adc_start( ADC );


  }


void RunADC()
{
  // ADC
  // configure Peripheral DMA
  Serial.println ("START adc...");
  
  begintime=micros();
  adcstartmicros=begintime;
  adctime = now();

 
  
  
  PDC_ADC->PERIPH_RPR = (uint32_t) adcbuffer; // address of buffer
  PDC_ADC->PERIPH_RCR = ADC_BUFFER_SIZE; 
  PDC_ADC->PERIPH_PTCR = PERIPH_PTCR_RXTEN; // enable receive
  while((adc_get_status(ADC) & ADC_ISR_ENDRX) == 0)
  {};



  
  endtime=micros();
  adcendmicros=endtime;
        Serial.print("STOP acd, done in ");
        Serial.print(endtime-begintime);
        Serial.print(" micros\n\n");  
        Serial.println("First channel is: ");
        Serial.print((adcbuffer[0] >> 12)& 0x000F);
        Serial.println();


// shift adcbuffer from random sequence to ch0,ch1,ch2,ch3,ch4,ch5,ch6,ch7,ch0,......

int firstchannel = (adcbuffer[0] >> 12)& 0x000F;
long temporaryadcterm;


begintime=micros();
// shift to rigth j elements

for (int j=0; j<firstchannel ; j++){

 // Serial.print(".");

  temporaryadcterm = adcbuffer[ADC_BUFFER_SIZE-1];
  for (int i=ADC_BUFFER_SIZE-1; i>0 ; i--){
    adcbuffer[i] = adcbuffer[i-1];
    }
  adcbuffer[0] = temporaryadcterm;
    
}
  endtime=micros();
Serial.print("transformation done in ");
Serial.print(endtime-begintime);
Serial.print(" micros\n\n");  


}


void ProcessADCBuffer()
{

int channelnummer, adcvalue;
long long acc_ch0;
long long acc_ch1;
long long acc_ch2;
long long acc_ch3;

long long i2t_temp;

// processing digital channels
for (int i=0; i < ADC_SAMPLES_PER_CHANNEL; i++){
  
  adcvalue = (adcbuffer[i*8+4])& 0x0FFF; // d ch0
  if (adcvalue < DSP_DIGITAL_TRESHOLD) digitalbuffer_ch0[i] = 0; else digitalbuffer_ch0[i] = 1;

  adcvalue = (adcbuffer[i*8+5])& 0x0FFF; // d ch1
  if (adcvalue < DSP_DIGITAL_TRESHOLD) digitalbuffer_ch1[i] = 0; else digitalbuffer_ch1[i] = 1;

    adcvalue = (adcbuffer[i*8+6])& 0x0FFF; // d ch2
  if (adcvalue < DSP_DIGITAL_TRESHOLD) digitalbuffer_ch2[i] = 0; else digitalbuffer_ch2[i] = 1;

    adcvalue = (adcbuffer[i*8+7])& 0x0FFF; // d ch3
  if (adcvalue < DSP_DIGITAL_TRESHOLD) digitalbuffer_ch3[i] = 0; else digitalbuffer_ch3[i] = 1;  
}

// processing analog channels

for (int i=0; i < ADC_SAMPLES_PER_CHANNEL; i++){  

  //  scaling and offsetting
  analogbuffer_ch0[i] = (adcbuffer[i*8+0])& 0x0FFF; // a ch0
  analogbuffer_ch0[i] = ( analogoffset[0] + analogbuffer_ch0[i] ) / analogscale [0];
  
  analogbuffer_ch1[i] = (adcbuffer[i*8+1])& 0x0FFF; // a ch1
  analogbuffer_ch1[i] = ( analogoffset[1] + analogbuffer_ch1[i] ) / analogscale [1];
  
  analogbuffer_ch2[i] = (adcbuffer[i*8+2])& 0x0FFF; // a ch2
  analogbuffer_ch2[i] = ( analogoffset[2] + analogbuffer_ch2[i] ) / analogscale [2];
  
  analogbuffer_ch3[i] = (adcbuffer[i*8+3])& 0x0FFF; // a ch3  
  analogbuffer_ch3[i] = ( analogoffset[3] + analogbuffer_ch3[i] ) / analogscale [3];
  
}
// filtering 

for (int i = ADC_SAMPLES_PER_CHANNEL-1; i >= DSP_FIR_FILTERTAPS-1; i--){ 
  acc_ch0 = 0;
  acc_ch1 = 0;
  acc_ch2 = 0;
  acc_ch3 = 0;
  
  for (int j = 0; j < DSP_FIR_FILTERTAPS; j++){ 
  acc_ch0 += analogfiltercoeff[j]*analogbuffer_ch0[i-j];   
  acc_ch1 += analogfiltercoeff[j]*analogbuffer_ch1[i-j];
 
  acc_ch2 += analogfiltercoeff[j]*analogbuffer_ch2[i-j];
  acc_ch3 += analogfiltercoeff[j]*analogbuffer_ch3[i-j];  
  }
  
  acc_ch0 = acc_ch0 / DSP_FIR_DCGAIN ;
  analogbuffer_ch0[i] = acc_ch0;

  //acc_ch1 = acc_ch1 / DSP_FIR_DCGAIN ;
  acc_ch1 = analogbuffer_ch1[i];
  analogbuffer_ch1[i] = acc_ch1;

  acc_ch2 = acc_ch2 / DSP_FIR_DCGAIN ;
  analogbuffer_ch2[i] = acc_ch2;

  acc_ch3 = acc_ch3 / DSP_FIR_DCGAIN ;
  analogbuffer_ch3[i] = acc_ch3;
}

// processing calculated values

  acc_ch0 = 0;
  acc_ch1 = 0;
  acc_ch2 = 0;
  acc_ch3 = 0;
  i2t_temp = 0;

for (int i=0; i < ADC_SAMPLES_PER_CHANNEL; i++){  
  
 // I2t calculation 0 channel - stored in c ch0
 i2t_temp = analogbuffer_ch0[i]*analogbuffer_ch0[i]*ADC_SAMPLETIME / 1000;
 if (digitalbuffer_ch0[i] > 0) acc_ch0 += i2t_temp;

  // I2t calculation 1 channel - stored in c ch1
 i2t_temp = analogbuffer_ch1[i]*analogbuffer_ch1[i]*ADC_SAMPLETIME / 1000;
 if (digitalbuffer_ch0[i] > 0) acc_ch1 += i2t_temp;

  // I2t calculation 2 channel - stored in c ch2
 i2t_temp = analogbuffer_ch2[i]*analogbuffer_ch2[i]*ADC_SAMPLETIME / 1000;
 if (digitalbuffer_ch0[i] > 0) acc_ch2 += i2t_temp;
 
  
  calculatedbuffer_ch0[i] = ( calcoffset[0] +acc_ch0 / 1000 ) / calcscale [0]; // c ch0
  calculatedbuffer_ch1[i] = ( calcoffset[1] +acc_ch1 / 1000 ) / calcscale [1];; // c ch1
  calculatedbuffer_ch2[i] = ( calcoffset[2] +acc_ch2 / 1000 ) / calcscale [2];; // c ch2
  calculatedbuffer_ch3[i] = 3; // c ch3  
}

  }



 /*-----------------------------------------------------------------
 * Ethernet functions 
  -----------------------------------------------------------------*/

 void InitEthernetSystem()
 {
  
 digitalWrite(10, LOW);                    //  turn on the W5100 chip!

  // start the Ethernet connection:
 if (useStaticIP == 1) {
  Ethernet.begin(mac, ip);
 } else
 {
  Serial.println("using DHCP server");
  if ( (Ethernet.begin(mac)) == 0) {
    Serial.println("Failed to configure Ethernet using DHCP");
    fatalBlink();
  }
 }
  // print your local IP address:
  Serial.print("My IP address: ");
  for (byte thisByte = 0; thisByte < 4; thisByte++) {
    // print the value of each byte of the IP address:
    Serial.print(Ethernet.localIP()[thisByte], DEC);
    Serial.print(".");
  }
  Serial.println();

  server.begin();
  Serial.print("server is at ");
  Serial.println(Ethernet.localIP());


 }




